import functools
import numpy as np

class q:
    from qlat_utils import (
            timer,
            timer_verbose,
            Coordinate,
            cache_call,
            set_zero,
            )
    from .c import (
            Geometry,
            FieldRealD,
            FieldComplexD,
            mk_fft,
            )

@q.cache_call(maxsize=16)
@q.timer
def mk_shift_xg_idx_arr(total_site, xg_shift):
    """
    only work without comm
    `total_site` and `xg_shift` should be tuple of ints.
    """
    total_site = np.array(total_site, dtype=np.int32)
    xg_shift = np.array(xg_shift, dtype=np.int32)
    total_volume = np.prod(total_site)
    xg_idx_arr = np.arange(total_volume).reshape(tuple(reversed(total_site))).T
    geo = q.Geometry(q.Coordinate(total_site))
    xg_arr = geo.xg_arr()
    xg_arr = (xg_arr + xg_shift) % total_site
    return xg_idx_arr[tuple(xg_arr.T)]

@q.timer
def smear_field_step_local(field, coef, n_steps=1):
    """
    Only work without comm
    Not very efficient
    """
    if n_steps == 0:
        return field.copy()
    geo = field.geo
    assert geo.num_node() == 1
    total_site = geo.total_site
    xg_shift_list = [
        [ 0, 0, 0, 1],
        [ 0, 0, 1, 0],
        [ 0, 1, 0, 0],
        [ 1, 0, 0, 0],
        [ 0, 0, 0, -1],
        [ 0, 0, -1, 0],
        [ 0, -1, 0, 0],
        [ -1, 0, 0, 0],
    ]
    n_dirs = len(xg_shift_list)
    xg_idx_arr_list = [ mk_shift_xg_idx_arr(tuple(total_site), tuple(xg_shift)) for xg_shift in xg_shift_list ]
    new_field = field.copy()
    coef_dir = coef / n_dirs / (1.0 - coef)
    for k in range(n_steps):
        for xg_idx_arr in xg_idx_arr_list:
            new_field[:] += coef_dir * field[xg_idx_arr]
        new_field *= 1.0 - coef
    return new_field

@q.timer
def smear_field_step(field, coef, n_steps=1):
    """
    Smear a density field
    Different from smearing the gauge field and measure the density.
    """
    if n_steps == 0:
        return field.copy()
    xg_shift_list = [
        [ 0, 0, 0, 1],
        [ 0, 0, 1, 0],
        [ 0, 1, 0, 0],
        [ 1, 0, 0, 0],
        [ 0, 0, 0, -1],
        [ 0, 0, -1, 0],
        [ 0, -1, 0, 0],
        [ -1, 0, 0, 0],
    ]
    xg_shift_list = [ q.Coordinate(xg) for xg in xg_shift_list ]
    n_dirs = len(xg_shift_list)
    coef_dir = coef / n_dirs / (1.0 - coef)
    new_field = field.copy()
    for k in range(n_steps):
        new_field *= 1 / coef_dir
        for xg_shift in xg_shift_list:
            f1 = field.shift(xg_shift)
            new_field += field.shift(xg_shift)
        new_field *= coef_dir * (1.0 - coef)
    return new_field

@q.cache_call(maxsize=128)
@q.timer_verbose
def mk_smear_mom_kernel(total_site, radius):
    r"""
    return f
    `radius` is the smear radius in lattice unit.
    `isinstance(f, q.FieldRealD)`
    `isinstance(total_site, tuple)`
    `f.total_site == q.Coordinate(total_site)`
    #
    f[:] == $G$
    #
    """
    assert isinstance(total_site, tuple)
    total_site = q.Coordinate(total_site)
    geo = q.Geometry(total_site)
    f = q.FieldRealD(geo, 1)
    xg_arr = geo.xg_arr
    q.set_zero(f)
    total_site_arr = total_site.to_numpy()
    xg_arr = geo.xg_arr()
    assert f[:, 0].shape == xg_arr[:, 0].shape
    if radius == np.inf:
        sel = True
        for i in range(4):
            sel = sel & (xg_arr[:, i] == 0)
        f[sel] = 1
    else:
        gg_arr = (2 * np.pi / total_site_arr / 2) * xg_arr[:, :]
        gg_arr = np.sum(np.sin(gg_arr)**2, axis=-1)
        gg_arr = (2 / 4 * radius**2) * gg_arr
        gg_arr = np.exp(-gg_arr)
        f[:] = gg_arr[:, None]
    return f

@q.cache_call(maxsize=128)
@q.timer_verbose
def mk_spatial_smear_mom_kernel(total_site, radius):
    r"""
    return f
    `radius` is the smear radius in lattice unit.
    `isinstance(f, q.FieldRealD)`
    `isinstance(total_site, tuple)`
    `f.total_site == q.Coordinate(total_site)`
    #
    f[:] == $G$
    #
    \ba
    G(\sigma, k) =
    \exp
    \Big(
    - \frac{2\sigma^2}{3} \sum_i \sin^2\big(\frac{k_i}{2}\big)
    \Big)
    \ea
    where k[i] = 2 pi * n[i] / total_site[i]
    """
    assert isinstance(total_site, tuple)
    total_site = q.Coordinate(total_site)
    geo = q.Geometry(total_site)
    f = q.FieldRealD(geo, 1)
    q.set_zero(f)
    total_site_arr = total_site.to_numpy()
    xg_arr = geo.xg_arr()
    if radius == np.inf:
        sel = True
        for i in range(3):
            sel = sel & (xg_arr[:, i] == 0)
        f[sel] = 1
    else:
        gg_arr = (2 * np.pi / total_site_arr[:3] / 2) * xg_arr[:, :3]
        gg_arr = np.sum(np.sin(gg_arr)**2, axis=-1)
        gg_arr = (2 / 3 * radius**2) * gg_arr
        gg_arr = np.exp(-gg_arr)
        f[:] = gg_arr[:, None]
    return f

@q.cache_call(maxsize=128)
@q.timer_verbose
def mk_sphere_sum_mom_kernel(total_site, radius, is_only_spatial):
    r"""
    return f
    `radius` is the smear radius in lattice unit.
    `isinstance(f, q.FieldRealD)`
    `isinstance(total_site, tuple)`
    `f.total_site == q.Coordinate(total_site)`
    #
    f[:] == $G$
    #
    """
    assert isinstance(total_site, tuple)
    total_site = q.Coordinate(total_site)
    geo = q.Geometry(total_site)
    f = q.FieldComplexD(geo, 1)
    q.set_zero(f)
    total_site_arr = total_site.to_numpy()
    if radius == np.inf:
        f[:] = 1
    else:
        if is_only_spatial:
            n_dim = 3
        else:
            n_dim = 4
        xg_arr = geo.xg_arr()
        assert f[:, 0].shape == xg_arr[:, 0].shape
        xg_rel_arr = np.empty_like(xg_arr)
        for mu in range(4):
            xg_rel_arr[:, mu] = np.where(xg_arr[:, mu] <= total_site_arr[mu] // 2, xg_arr[:, mu], xg_arr[:, mu] - total_site_arr[mu])
        dis_sqr_arr = 0
        for mu in range(n_dim):
            dis_sqr_arr += xg_rel_arr[:, mu]**2
        sel = dis_sqr_arr < radius**2
        f[sel] = 1
    mode_fft = 1
    fft_f = q.mk_fft(True, is_normalizing=False, is_only_spatial=is_only_spatial, mode_fft=mode_fft)
    f_f = fft_f * f
    f = q.FieldRealD(geo, 1)
    f[:] = f_f[:].real
    return f

@q.timer
def sphere_sum_field(field, radius, *, is_only_spatial=False):
    r"""
    return sphere_summed_field
    field must at least be complex type.
    #
    $$
    \ba
    f_\text{sphere-summed}(x)
    \approx
    \sum_{y} \theta(|x-y| < r) f(y)
    \ea
    $$
    """
    total_site = field.geo.total_site
    fk = mk_sphere_sum_mom_kernel(total_site.to_tuple(), radius, is_only_spatial)
    fft1 = q.mk_fft(True, is_only_spatial=is_only_spatial, is_normalizing=True)
    fft2 = q.mk_fft(False, is_only_spatial=is_only_spatial, is_normalizing=True)
    summed_field = field.copy()
    ftmp = fft1 * field
    ftmp *= fk
    ftmp = fft2 * ftmp
    summed_field @= ftmp
    return summed_field

@q.timer
def smear_field(field, radius, *, is_only_spatial=False):
    r"""
    return smeared_field
    field must at least be complex type.
    #
    $$
    \ba
    f_\text{smear}(x)
    \approx
    \frac{
    \sum_y f(y) \exp( - (x - y)^2 / (2 r^2) )
    }{
    \sum_y \exp( - y^2 / (2 r^2) )
    }
    \ea
    $$
    """
    total_site = field.geo.total_site
    if is_only_spatial:
        fk = mk_spatial_smear_mom_kernel(total_site.to_tuple(), radius)
    else:
        fk = mk_smear_mom_kernel(total_site.to_tuple(), radius)
    fft1 = q.mk_fft(True, is_only_spatial=is_only_spatial, is_normalizing=True)
    fft2 = q.mk_fft(False, is_only_spatial=is_only_spatial, is_normalizing=True)
    smeared_field = field.copy()
    ftmp = fft1 * field
    ftmp *= fk
    ftmp = fft2 * ftmp
    smeared_field @= ftmp
    return smeared_field

@q.timer
def field_convolution(f1, f2, idx1=None, idx2=None, *, is_only_spatial=False):
    r"""
    return ff
    where
    ff[xg_rel] = \sum_{xg} f2[xg + xg_rel, idx2] * f1[xg, idx1]
    #
    isinstance(f1, q.FieldComplexD)
    isinstance(f1, q.FieldComplexD)
    len(idx1) == len(idx2)
    """
    assert isinstance(f1, q.FieldComplexD)
    assert isinstance(f2, q.FieldComplexD)
    geo = f1.geo
    assert geo.total_site == f2.geo.total_site
    if idx1 is None and idx2 is None:
        assert f1.multiplicity == f2.multiplicity
        idx1 = np.arange(f1.multiplicity, dtype=np.int32)
        idx2 = np.arange(f2.multiplicity, dtype=np.int32)
    assert len(idx1) == len(idx2)
    assert np.all(idx1 < f1.multiplicity)
    assert np.all(idx2 < f2.multiplicity)
    mode_fft = 1
    fft_f = q.mk_fft(True, is_normalizing=False, is_only_spatial=is_only_spatial, mode_fft=mode_fft)
    fft_b = q.mk_fft(False, is_normalizing=False, is_only_spatial=is_only_spatial, mode_fft=mode_fft)
    f_f2 = fft_f * f2
    f_f1 = fft_b * f1
    f_ff = q.FieldComplexD(geo, len(idx1))
    f_ff[:] = f_f2[:, idx2] * f_f1[:, idx1]
    if is_only_spatial:
        f_ff *= geo.total_site[3] / geo.total_volume
    else:
        f_ff *= 1.0 / geo.total_volume
    ff = fft_b * f_ff
    assert ff.multiplicity == len(idx1)
    return ff



# Author: Luchang Jin 2023

import qlat as q
import sys

if len(sys.argv) < 2:
    q.displayln_info("Usage: fields-rewrite [--force] path1 path2 ...")
    q.displayln_info("Only rewrite if there is duplicates unless --force is given.")
    exit()

q.begin_with_mpi()

argv = sys.argv[1:]

is_force = q.get_option("--force", argv=argv, is_removing_from_argv=True)

path_list = argv

for path in path_list:
    path = q.remove_trailing_slashes(path)
    q.displayln_info(-1, f"INFO: Consider '{path}'.")
    sfr = q.open_fields(path, "r")
    has_duplicates = sfr.has_duplicates()
    if has_duplicates:
        q.displayln_info(-1, f"INFO: '{path}' has_duplicates.")
    do_rewrite = has_duplicates or is_force
    if do_rewrite:
        new_path = path + ".rewrite.acc"
        q.displayln_info(-1, f"INFO: '{path}' start to rewrite to '{new_path}'.")
        sfw = q.open_fields(new_path, "w", sfr.new_size_node())
        tags = sfr.list()
        for tag in tags:
            if tag in sfw:
                q.displayln_info(-1, f"INFO: Skip duplicated '{tag}' of '{sfr.path()}'.")
                continue
            q.displayln_info(-1, f"INFO: Read '{tag}' of '{sfr.path()}'.")
            obj = sfr.read_as_char(tag)
            q.displayln_info(-1, f"INFO: Write '{tag}' of '{sfw.path()}'.")
            obj.save_direct(sfw, tag)
        sfw.close()
    sfr.close()
    if do_rewrite:
        bak_path = path + ".bak"
        q.displayln_info(-1, f"INFO: Rename '{path}' to '{bak_path}'.")
        q.qrename_info(path, bak_path)
        q.displayln_info(-1, f"INFO: Rename '{new_path}' to '{path}'.")
        q.qrename_info(new_path, path)
    q.displayln_info(-1, f"INFO: Done '{path}'.")

q.timer_display()

q.end_with_mpi()

exit()



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils.all cimport *
from . cimport everything as cc

import qlat_utils as q

cdef class GaugeAction:

    def __cinit__(self):
        self.xx = cc.GaugeAction()
        self.cdata = <cc.Long>&(self.xx)

    def __init__(self, cc.RealD beta, cc.RealD c1=0.0):
        self.xx.beta = beta
        self.xx.c1 = c1

    def __imatmul__(self, GaugeAction v1):
        cc.assign_direct(self.xx, v1.xx)
        return self

    def beta(self):
        return self.xx.beta

    def c1(self):
        return self.xx.c1



r"""

# Plaq dependent flow\n
Standard Wilson flow action:
$$
S_\mathrm{Wilson} = \frac{\beta}{2}\sum_{x,\mu,\nu} \Big(1 - \frac{1}{3}\mathrm{Re}\mathrm{Tr} U_{\mu,\nu}\Big)
$$
with $\beta = 3$.\n
Modified flow to make the slope depends on the plaq value:
$$
S_f = -\frac{\beta}{2}\sum_{x,\mu,\nu} f\big(\frac{1}{3}\mathrm{Re}\mathrm{Tr} U_{\mu,\nu}\big)
$$\n
We can choose the function $f$ that smooths gauge field and therefore fix the total topological charge:
$$
\frac{d}{dp}f_\mathrm{Freeze}(p)
=
1 - p
$$
Note that $f_\mathrm{Freeze}$ suppresses plaq significantly deviate from $1$ and therefore prevent topological charge tunnelling during flow.\n
We can also choose the function $f$ that shrinks instanton:
$$
\frac{d}{dp}f_\mathrm{Shrink}(p)
=
\frac{\epsilon}{1 - p + \epsilon} + b
$$
with a possible choice of parameters be $\epsilon = 0.005$ and $b=0.5$.
Note that $f_\mathrm{Shrink}$ tends to shrink the size of instanton therefore enhance topological charge tunnelling during flow.\n
We can also choose the function $f$ that shrinks large instanton but prevent small instanton from tunnelling:
$$
\frac{d}{dp}f_\mathrm{Localize}(p)
=
\frac{\epsilon}{1 - p + \epsilon} + b (1 - p)
$$
with a possible choice of parameters be $\epsilon = 0.002$ and $b=50$.\n
We can also choose the function $f$ that mimic Wilson flow but prevent small instanton from tunnelling:
$$
\frac{d}{dp}f_\mathrm{Preserve}(p)
=
max(1, b (1 - p))
$$
with a possible choice of parameters be $b=50$.

"""

# --------------------------------------

import numpy as np
from pprint import pformat


class q:
    from qlat_utils import (
        timer,
        Coordinate,
        smod_coordinate_d,
        get_fname,
        RngState,
        parallel_map,
        random_permute,
        displayln_info,
        get_num_node,
        is_test,
        json_results_append,
        save_pickle_obj,
        save_json_obj,
        filter_np_results,
    )
    from qlat.c import (
        Geometry,
        PointsSelection,
        FieldComplexD,
        FieldRealD,
        gf_plaq_flow_force,
        gf_evolve,
        gf_plaq_field,
        gf_topology_field,
        gf_topology,
        Coordinate,
        CoordinateD,
        PointsSelection,
        gf_reduce_half,
        gf_wilson_flow_step,
        gf_energy_density_field,
        gf_topology_field_clf,
        gf_topology_field,
        gf_topology_terms,
        gf_plaq_action_density_field,
    )
    from .field_analysis import (
        smear_field,
        sphere_sum_field,
    )
    from .mpi_utils import (
        get_mpi_chunk,
        get_comm,
    )

# --------------------------------------


@q.timer
def gf_flow_topo(
    gf, step_size,
    flow_type=None,
    wilson_flow_integrator_type=None,
):
    """
    return step_size / norm
    Modify ``gf`` in place.
    Default Wilson flow with Euler integrator.
    ```
    flow_type in [
        None,
        "Wilson", # c1 = 0.0
        "Iwasaki", # c1 = -0.331
        "DBW2", # c1 = -1.4008
        "Shrink",
        "Freeze",
        "Localize",
        c1,
    ]
    wilson_flow_integrator_type in [
        None,
        "euler", # Unlike ``q.gf_wilson_flow_step``, the default here is "euler"
        "runge-kutta",
    ]
    ```
    c1 = 0.0 # Wilson
    c1 = -0.331 # Iwasaki
    c1 = -1.4008 # DBW2
    """
    fname = q.get_fname()
    geo = gf.geo
    if flow_type is None:
        flow_type = "Wilson"
    if wilson_flow_integrator_type is None:
        wilson_flow_integrator_type = "euler"
    norm = 1
    if isinstance(flow_type, (int, float,)):
        c1 = flow_type
        q.gf_wilson_flow_step(
            gf, step_size, c1=c1, wilson_flow_integrator_type=wilson_flow_integrator_type,
        )
    elif flow_type in ["Wilson", "Iwasaki", "DBW2", ]:
        if flow_type == "Wilson":
            c1 = 0.0
        elif flow_type == "Iwasaki":
            c1 = -0.331
        elif flow_type == "DBW2":
            c1 = -1.4008
        else:
            raise Exception(f"{fname}: {flow_type=}")
        q.gf_wilson_flow_step(
            gf, step_size, c1=c1, wilson_flow_integrator_type=wilson_flow_integrator_type,
        )
    else:
        plaq_factor = q.FieldRealD(geo, 6)
        f_plaq = q.gf_plaq_field(gf)
        plaq_min = np.min(f_plaq.glb_min()[:]).item()
        plaq_max = np.max(f_plaq.glb_max()[:]).item()
        if flow_type == "Shrink":
            eps = 0.005
            base = 0.5
            norm = eps / (1 - plaq_max + eps) + base
            plaq_factor[:] = eps / (1 - f_plaq[:] + eps) + base
        elif flow_type == "Freeze":
            norm = 1 - plaq_min
            plaq_factor[:] = 1 - f_plaq[:]
        elif flow_type == "Localize":
            eps = 0.002
            base = 50
            norm = max(
                eps / (1 - plaq_max + eps) + (base * (1 - plaq_max)),
                eps / (1 - plaq_min + eps) + (base * (1 - plaq_min)),
            )
            plaq_factor[:] = eps / (1 - f_plaq[:] + eps) + \
                (base * (1 - f_plaq[:]))
        elif flow_type == "Preserve":
            base = 50
            norm = max(1, base * (1 - plaq_min))
            plaq_factor[:] = np.maximum(1, base * (1 - f_plaq[:]))
        else:
            raise Exception(f"{fname}: {flow_type=}")
        gm_force = q.gf_plaq_flow_force(gf, plaq_factor)
        q.gf_evolve(gf, gm_force, step_size / norm)
    return q.filter_np_results(step_size / norm)


@q.timer
def gf_energy_derivative_density_field_topo(
    gf,
    *,
    epsilon=0.0125,
    flow_typw=None,
    wilson_flow_integrator_type=None,
):
    """
    flow_type in [
        None,
        "Wilson", # c1 = 0.0
        "Iwasaki", # c1 = -0.331
        "DBW2", # c1 = -1.4008
        c1,
    ]
    wilson_flow_integrator_type in [
        None,
        "euler", # Unlike ``q.gf_wilson_flow_step``, the default here is "euler"
        "runge-kutta",
    ]
    """
    gf1 = gf.copy()
    gf_flow_topo(gf1, epsilon, flow_typw, wilson_flow_integrator_type)
    fd1 = q.gf_energy_density_field(gf1)
    gf1 @= gf
    gf_flow_topo(gf1, -epsilon, flow_typw, wilson_flow_integrator_type)
    fd2 = q.gf_energy_density_field(gf1)
    fd1 -= fd2
    fd1 *= 1 / (2 * epsilon)
    return fd1


@q.timer
def mk_plaq_xg_arr(geo):
    n_plaq = 6
    xg_arr = geo.xg_arr()
    (n_points, n_dim,) = xg_arr.shape
    assert n_dim == 4
    assert n_points == geo.local_volume
    plaq_xg_arr = np.zeros((n_points, n_plaq, n_dim,), dtype=np.float64)
    plaq_xg_arr[:, :, :] = xg_arr[:, None, :]
    plaq_xg_arr[:, 0, :] += [0.5, 0.5, 0, 0, ]
    plaq_xg_arr[:, 1, :] += [0.5, 0, 0.5, 0, ]
    plaq_xg_arr[:, 2, :] += [0.5, 0, 0, 0.5, ]
    plaq_xg_arr[:, 3, :] += [0, 0.5, 0.5, 0, ]
    plaq_xg_arr[:, 4, :] += [0, 0.5, 0, 0.5, ]
    plaq_xg_arr[:, 5, :] += [0, 0, 0.5, 0.5, ]
    # print(plaq_xg_arr.shape)
    return plaq_xg_arr


@q.timer
def get_extreme_plaq_xg_list(plaq_xg_arr, f_plaq, threshold):
    """
    return extreme_plaq_xg_list
    extreme_plaq_xg_list = [ (plaq, xg,), ... ]
    #
    e.g.
    plaq_xg_arr = mk_plaq_xg_arr(geo)
    f_plaq = q.gf_plaq_field(gf)
    threshold = 0.1
    extreme_plaq_xg_list = get_extreme_plaq_xg_list(plaq_xg_arr, f_plaq, threshold)
    """
    sel = f_plaq[:] < (1 - threshold)
    plaq_arr = f_plaq[sel]
    extreme_xg_arr = plaq_xg_arr[sel]
    extreme_plaq_xg_list = list(
        zip(plaq_arr.tolist(), extreme_xg_arr.tolist()))
    all_extreme_plaq_xg_list = q.get_comm().allgather(extreme_plaq_xg_list)
    extreme_plaq_xg_list = []
    for sub in all_extreme_plaq_xg_list:
        extreme_plaq_xg_list += sub
    extreme_plaq_xg_list.sort()
    return extreme_plaq_xg_list


def point_d_dis_sqr(x, y, total_site):
    return q.smod_coordinate_d(x - y, total_site).sqr()


@q.timer
def get_group_extreme_plaq_xg_list(extreme_plaq_xg_list, total_site, dis_sqr_limit):
    """
    Group points with distance square less than `dis_sqr_limit`.
    #
    return group_extreme_plaq_xg_list
    group_extreme_plaq_xg_list = [
        [ (plaq, xg, dis_sq,), ... ],
        ...
    ]
    #
    e.g.
    extreme_plaq_xg_list = get_extreme_plaq_xg_list(plaq_xg_arr, f_plaq, threshold)
    total_site = q.Coordinate([ 4, 4, 4, 8, ])
    dis_sqr_limit = 5
    group_extreme_plaq_xg_list = get_group_extreme_plaq_xg_list(extreme_plaq_xg_list, total_site, dis_sqr_limit)
    """
    total_site = q.CoordinateD(total_site)
    group_extreme_plaq_xg_list = []
    for plaq, xg in extreme_plaq_xg_list:
        x1 = q.CoordinateD(xg)
        b = False
        for l in group_extreme_plaq_xg_list:
            assert len(l) >= 1
            x2 = q.CoordinateD(l[0][1])
            dis_sq = point_d_dis_sqr(x1, x2, total_site)
            if dis_sq < dis_sqr_limit:
                l.append((plaq, xg, dis_sq))
                b = True
                break
        if not b:
            l = [(plaq, xg, 0,), ]
            group_extreme_plaq_xg_list.append(l)
    return group_extreme_plaq_xg_list


@q.timer
def process_inst_list(inst_list):
    """
    inst_list[inst_idx] = [
        dict(flow_time, current_spacing, total_site, o_xg_d, o_xg, xg_d, xg, plaq, s_topo),
        ...
    ]
    #
    p_inst_list = process_inst_list(inst_list)
    #
    p_inst_list[inst_idx] = dict(
        # original `inst_idx` for `inst_list`
        inst_idx,
        inst_raw = inst_list[inst_idx],
        #
        # Values are usually taken at minimum `plaq`
        o_xg_d,
        o_total_site,
        xg_d,
        total_site,
        current_spacing,
        plaq,
        flow_time,
        num_plaq,
        dis_sqr_max,
        #
        flow_time_start,
        flow_time_end,
        dis_sqr_max_max,
        c_dis_sqr_max,
        #
        plaq_list,
        s_topo_list,
        num_plaq_list,
        dis_sqr_avg_list,
        dis_sqr_max_list,
        c_dis_sqr_list,
        delta_s_topo,
    )
    """
    p_inst_list = []
    for inst_idx, inst in enumerate(inst_list):
        s_topo_list = []
        plaq_list = []
        num_plaq_list = []
        dis_sqr_avg_list = []
        dis_sqr_max_list = []
        c_dis_sqr_list = []
        plaq = 1.0
        for info_dict in inst:
            plaq_list.append(info_dict["plaq"])
            num_plaq_list.append(info_dict["num_plaq"])
            dis_sqr_avg_list.append(info_dict["dis_sqr_avg"])
            dis_sqr_max_list.append(info_dict["dis_sqr_max"])
            if info_dict["s_topo"] is not None:
                s_topo_list.append(info_dict["s_topo"])
            if info_dict["plaq"] < plaq:
                plaq = info_dict["plaq"]
                flow_time = info_dict["flow_time"]
                s_topo = info_dict["s_topo"]
                num_plaq = info_dict["num_plaq"]
                dis_sqr_avg = info_dict["dis_sqr_avg"]
                dis_sqr_max = info_dict["dis_sqr_max"]
                o_xg_d = info_dict["o_xg_d"]
                xg_d = info_dict["xg_d"]
        for info_dict in inst:
            total_site_d = q.CoordinateD(info_dict["total_site"])
            c_dis_sqr = point_d_dis_sqr(q.CoordinateD(
                xg_d), q.CoordinateD(info_dict["xg_d"]), total_site_d)
            c_dis_sqr_list.append(c_dis_sqr)
        s_topo_arr = np.array(s_topo_list, dtype=np.float64)
        s_topo_list = s_topo_arr.T.tolist()
        delta_s_topo = None
        if len(s_topo_arr) > 0:
            delta_s_topo = (s_topo_arr[0] - s_topo_arr[-1]).tolist()
        total_site = inst[0]["total_site"]
        current_spacing = inst[0]["current_spacing"]
        topo_sphere_sum_radius_list = inst[0]["topo_sphere_sum_radius_list"]
        dis_sqr_max_max = np.max(dis_sqr_max_list).item()
        c_dis_sqr_max = np.max(c_dis_sqr_list).item()
        p_inst = dict()
        p_inst["inst_idx"] = inst_idx
        p_inst["inst_raw"] = inst
        p_inst["o_xg_d"] = o_xg_d
        p_inst["o_total_site"] = [current_spacing * c for c in total_site]
        p_inst["xg_d"] = xg_d
        p_inst["total_site"] = total_site
        p_inst["current_spacing"] = current_spacing
        p_inst["plaq"] = plaq
        p_inst["flow_time"] = flow_time
        p_inst["flow_time_start"] = inst[0]["flow_time"]
        p_inst["flow_time_end"] = inst[-1]["flow_time"]
        p_inst["num_plaq"] = num_plaq
        p_inst["dis_sqr_avg"] = dis_sqr_avg
        p_inst["dis_sqr_max"] = dis_sqr_max
        p_inst["dis_sqr_max_max"] = dis_sqr_max_max
        p_inst["c_dis_sqr_max"] = c_dis_sqr_max
        p_inst["delta_s_topo"] = delta_s_topo
        p_inst["plaq_list"] = plaq_list
        p_inst["s_topo_list"] = s_topo_list
        p_inst["topo_sphere_sum_radius_list"] = topo_sphere_sum_radius_list
        p_inst["num_plaq_list"] = num_plaq_list
        p_inst["dis_sqr_avg_list"] = dis_sqr_avg_list
        p_inst["dis_sqr_max_list"] = dis_sqr_max_list
        p_inst["c_dis_sqr_list"] = c_dis_sqr_list
        p_inst_list.append(p_inst)

    def mk_closest_inst_info(inst_dis_sqr, p_inst):
        """
        e.g.
        mk_closest_inst_info(inst_dis_sqr, p_inst_2)
        """
        closest_inst_info = dict()
        closest_inst_info["inst_dis_sqr"] = inst_dis_sqr
        closest_inst_info["current_spacing"] = p_inst["current_spacing"]
        closest_inst_info["inst_idx"] = p_inst["inst_idx"]
        closest_inst_info["o_xg_d"] = p_inst["o_xg_d"]
        closest_inst_info["plaq"] = p_inst["plaq"]
        closest_inst_info["flow_time"] = p_inst["flow_time"]
        closest_inst_info["flow_time_start"] = p_inst["flow_time_start"]
        closest_inst_info["flow_time_end"] = p_inst["flow_time_end"]
        closest_inst_info["dis_sqr_max_max"] = p_inst["dis_sqr_max_max"]
        closest_inst_info["c_dis_sqr_max"] = p_inst["c_dis_sqr_max"]
        closest_inst_info["delta_s_topo"] = p_inst["delta_s_topo"]
        return closest_inst_info
    for p_inst in p_inst_list:
        inst_idx = p_inst["inst_idx"]
        o_xg_d = p_inst["o_xg_d"]
        o_xg_d = q.CoordinateD(o_xg_d)
        o_total_site = p_inst["o_total_site"]
        o_total_site_d = q.CoordinateD(o_total_site)
        current_spacing = p_inst["current_spacing"]
        flow_time = p_inst["flow_time"]
        flow_time_start = p_inst["flow_time_start"]
        flow_time_end = p_inst["flow_time_end"]
        closest_inst_dis_sqr = None
        closest_inst_info = None
        closest_sim_inst_dis_sqr = None
        closest_sim_inst_info = None
        for p_inst_2 in p_inst_list:
            inst_idx_2 = p_inst_2["inst_idx"]
            if inst_idx_2 == inst_idx:
                continue
            o_xg_d_2 = p_inst_2["o_xg_d"]
            o_xg_d_2 = q.CoordinateD(o_xg_d_2)
            o_total_site_2 = p_inst_2["o_total_site"]
            assert o_total_site_2 == o_total_site
            current_spacing_2 = p_inst_2["current_spacing"]
            flow_time_start_2 = p_inst_2["flow_time_start"]
            flow_time_end_2 = p_inst_2["flow_time_end"]
            inst_dis_sqr = point_d_dis_sqr(o_xg_d, o_xg_d_2, o_total_site_d)
            if (closest_inst_dis_sqr is None) or (inst_dis_sqr < closest_inst_dis_sqr):
                closest_inst_dis_sqr = inst_dis_sqr
                closest_inst_info = mk_closest_inst_info(
                    inst_dis_sqr, p_inst_2)
            if ((flow_time_start <= flow_time_start_2 <= flow_time_end)
                    or (flow_time_start <= flow_time_end_2 <= flow_time_end)
                    or (flow_time_start_2 <= flow_time_start <= flow_time_end_2)
                    or (flow_time_start_2 <= flow_time_end <= flow_time_end_2)
                ):
                if (closest_sim_inst_dis_sqr is None) or (inst_dis_sqr < closest_sim_inst_dis_sqr):
                    assert current_spacing_2 == current_spacing
                    closest_sim_inst_dis_sqr = inst_dis_sqr
                    closest_sim_inst_info = mk_closest_inst_info(
                        inst_dis_sqr, p_inst_2)
        p_inst["closest_sim_inst_info"] = closest_sim_inst_info
        p_inst["closest_inst_info"] = closest_inst_info
    for p_inst in p_inst_list:
        delta_s_topo = p_inst["delta_s_topo"]
        estimate_topo_charge = 0
        p_inst["estimate_topo_charge"] = estimate_topo_charge
        if delta_s_topo is None:
            continue
        delta_s_topo = np.array(delta_s_topo, dtype=np.float64)
        if len(delta_s_topo) > 1:
            estimate_topo_charge = (
                delta_s_topo[0] - (delta_s_topo[1] - 1.5 * delta_s_topo[0]) / 3) / 0.44
        elif len(delta_s_topo) > 0:
            estimate_topo_charge = delta_s_topo[0] / 0.44
        p_inst["estimate_topo_charge"] = estimate_topo_charge
        #
        # # If there is another instanton very close, consider the two instantons together. (Disable now)
        # p_inst["estimate_topo_charge_orig"] = None
        # inst_idx = p_inst["inst_idx"]
        # current_spacing = p_inst["current_spacing"]
        # closest_sim_inst_info = p_inst["closest_sim_inst_info"]
        # if closest_sim_inst_info is None:
        #     continue
        # inst_dis_sqr_2 = closest_sim_inst_info["inst_dis_sqr"]
        # pairwise_inst_dis_sqr_max_limit = 4**2
        # pairwise_inst_dis_sqr_min_limit = 5
        # if inst_dis_sqr_2 >= pairwise_inst_dis_sqr_max_limit * current_spacing**2:
        #     continue
        # if inst_dis_sqr_2 <= pairwise_inst_dis_sqr_min_limit * current_spacing**2:
        #     continue
        # inst_idx_2 = closest_sim_inst_info["inst_idx"]
        # closest_sim_inst_info_2 = p_inst_list[inst_idx_2]["closest_sim_inst_info"]
        # assert closest_sim_inst_info_2 is not None
        # if closest_sim_inst_info_2["inst_idx"] != inst_idx:
        #     continue
        # delta_s_topo_2 = closest_sim_inst_info["delta_s_topo"]
        # if delta_s_topo_2 is None:
        #     continue
        # delta_s_topo_2 = np.array(delta_s_topo_2, dtype=np.float64)
        # assert len(delta_s_topo_2) == len(delta_s_topo)
        # delta_s_topo_avg = (delta_s_topo + delta_s_topo_2) / 2
        # delta_s_topo_diff = (delta_s_topo - delta_s_topo_2) / 2
        # if len(delta_s_topo) > 1:
        #     estimate_topo_charge = round((delta_s_topo_avg[0] - (delta_s_topo_avg[1] - 1.5 * delta_s_topo_avg[0]) / 3) / 0.44)
        #     estimate_topo_charge += round((delta_s_topo_diff[0] - (delta_s_topo_diff[1] - 1.5 * delta_s_topo_diff[0]) / 3) / 0.40)
        # elif len(delta_s_topo) > 0:
        #     estimate_topo_charge = round(delta_s_topo_avg[0] / 0.44)
        #     estimate_topo_charge += round(delta_s_topo_diff[0] / 0.40)
        # p_inst["estimate_topo_charge_orig"] = p_inst["estimate_topo_charge"]
        # p_inst["estimate_topo_charge"] = estimate_topo_charge
    # p_inst_list.sort(key=lambda v: v["flow_time"])
    return p_inst_list


@q.timer
def displayln_info_topo_info(topo_info):
    """
    `topo_info` is a `dict`
    """
    info = topo_info
    flow_time = info["flow_time"]
    step_counter = info["step_counter"]
    flow_type = info["flow_type"]
    current_spacing = info["current_spacing"]
    plaq = info["plaq"]
    plaq_min = info["plaq_min"]
    num_instanton = info["num_instanton"]
    topo = info["topo"]
    q.displayln_info(0, f"{flow_type} flow_time={flow_time:9.2f} ; topo={topo:10.3f} ; plaq_min={plaq_min:8.5f} ; plaq={plaq:9.6f} ; num_instanton={num_instanton:11.2f} ; step_counter={step_counter:5} ; current_spacing={current_spacing}")


@q.timer
def displayln_info_p_inst(p_inst):
    """
    e.g.:
    p_inst_list = process_inst_list(inst_list)
    for p_inst in p_inst_list:
        displayln_info_p_inst(p_inst)
    """
    inst_idx = p_inst["inst_idx"]
    inst_raw = p_inst["inst_raw"]
    o_xg_d = p_inst["o_xg_d"]
    o_total_site = p_inst["o_total_site"]
    xg_d = p_inst["xg_d"]
    total_site = p_inst["total_site"]
    current_spacing = p_inst["current_spacing"]
    plaq = p_inst["plaq"]
    flow_time = p_inst["flow_time"]
    flow_time_start = p_inst["flow_time_start"]
    flow_time_end = p_inst["flow_time_end"]
    num_plaq = p_inst["num_plaq"]
    dis_sqr_avg = p_inst["dis_sqr_avg"]
    dis_sqr_max = p_inst["dis_sqr_max"]
    dis_sqr_max_max = p_inst["dis_sqr_max_max"]
    c_dis_sqr_max = p_inst["c_dis_sqr_max"]
    delta_s_topo = p_inst["delta_s_topo"]
    topo_sphere_sum_radius_list = p_inst["topo_sphere_sum_radius_list"]
    estimate_topo_charge = p_inst["estimate_topo_charge"]
    plaq_list = p_inst["plaq_list"]
    s_topo_list = p_inst["s_topo_list"]
    num_plaq_list = p_inst["num_plaq_list"]
    dis_sqr_avg_list = p_inst["dis_sqr_avg_list"]
    dis_sqr_max_list = p_inst["dis_sqr_max_list"]
    c_dis_sqr_list = p_inst["c_dis_sqr_list"]
    closest_sim_inst_info = p_inst["closest_sim_inst_info"]
    closest_inst_info = p_inst["closest_inst_info"]
    idx = inst_idx
    q.displayln_info(0, f"{idx:3}:")
    q.displayln_info(0, f"{idx:3}: o_xg_d={o_xg_d}")
    q.displayln_info(0, f"{idx:3}: o_total_site={o_total_site}")
    q.displayln_info(0, f"{idx:3}: xg_d={xg_d}")
    q.displayln_info(0, f"{idx:3}: total_site={total_site}")
    q.displayln_info(0, f"{idx:3}: current_spacing={current_spacing}")
    q.displayln_info(0, f"{idx:3}: plaq={plaq:.4f}")
    q.displayln_info(0, f"{idx:3}: flow_time={flow_time:.2f}")
    q.displayln_info(0, f"{idx:3}: num_plaq={num_plaq}")
    # q.displayln_info(0, f"{idx:3}: dis_sqr_avg={dis_sqr_avg}")
    q.displayln_info(0, f"{idx:3}: dis_sqr_max={dis_sqr_max}")
    q.displayln_info(0, f"{idx:3}: dis_sqr_max_max={dis_sqr_max_max}")
    q.displayln_info(0, f"{idx:3}: c_dis_sqr_max={c_dis_sqr_max}")
    q.displayln_info(0, f"{idx:3}: flow_time_start={flow_time_start:.2f}")
    q.displayln_info(0, f"{idx:3}: flow_time_end={flow_time_end:.2f}")
    q.displayln_info(
        0, f"{idx:3}: plaq_list={[f'{v:.4f}' for v in plaq_list]}")
    q.displayln_info(
        0, f"{idx:3}: s_topo_list={[[f'{v:.3f}' for v in l] for l in s_topo_list]}")
    q.displayln_info(
        0, f"{idx:3}: topo_sphere_sum_radius_list={topo_sphere_sum_radius_list}")
    q.displayln_info(0, f"{idx:3}: num_plaq_list={num_plaq_list}")
    q.displayln_info(
        0, f"{idx:3}: dis_sqr_avg_list={[f'{v:.3f}' for v in dis_sqr_avg_list]}")
    q.displayln_info(0, f"{idx:3}: dis_sqr_max_list={dis_sqr_max_list}")
    q.displayln_info(0, f"{idx:3}: c_dis_sqr_list={c_dis_sqr_list}")
    q.displayln_info(
        0, f"{idx:3}: closest_sim_inst_info={closest_sim_inst_info}")
    q.displayln_info(0, f"{idx:3}: closest_inst_info={closest_inst_info}")
    if delta_s_topo is None:
        q.displayln_info(0, f"{idx:3}: delta_s_topo={delta_s_topo}")
    else:
        q.displayln_info(
            0, f"{idx:3}: delta_s_topo={[f'{v:.3f}' for v in delta_s_topo]}")
    q.displayln_info(
        0, f"{idx:3}: estimate_topo_charge={estimate_topo_charge}")
    # estimate_topo_charge_orig = p_inst["estimate_topo_charge_orig"]
    # if estimate_topo_charge_orig is not None:
    #     q.displayln_info(0, f"{idx:3}: estimate_topo_charge_orig={estimate_topo_charge_orig}")


@q.timer
def displayln_info_p_inst_list(p_inst_list):
    """
    e.g.:
    p_inst_list = process_inst_list(inst_list)
    displayln_info_p_inst_list(p_inst_list)
    """
    tot_topo = 0
    tot_inst = 0
    tot_topo_dict = dict()
    tot_inst_dict = dict()
    for p_inst in p_inst_list:
        displayln_info_p_inst(p_inst)
        inst_idx = p_inst["inst_idx"]
        estimate_topo_charge = p_inst["estimate_topo_charge"]
        current_spacing = p_inst["current_spacing"]
        idx = inst_idx
        tot_topo += estimate_topo_charge
        tot_inst += abs(estimate_topo_charge)
        q.displayln_info(0, f"{idx:3}: tot_inst={tot_inst}")
        q.displayln_info(0, f"{idx:3}: tot_topo={tot_topo}")
        if current_spacing in tot_topo_dict:
            tot_topo_dict[current_spacing] += estimate_topo_charge
            tot_inst_dict[current_spacing] += abs(estimate_topo_charge)
        else:
            tot_topo_dict[current_spacing] = estimate_topo_charge
            tot_inst_dict[current_spacing] = abs(estimate_topo_charge)
    q.displayln_info(0, f"Final:")
    q.displayln_info(0, f"Final: tot_inst_dict={tot_inst_dict}")
    q.displayln_info(0, f"Final: tot_topo_dict={tot_topo_dict}")
    q.displayln_info(0, f"Final: tot_inst={tot_inst}")
    q.displayln_info(0, f"Final: tot_topo={tot_topo}")


def get_tot_topo_count(p_inst_list, flow_time):
    """
    Get the total topological charge by counting the number of instanton minus anti-instanton.
    """
    topo_count = sum([round(p_inst["estimate_topo_charge"])
                     for p_inst in p_inst_list if p_inst["flow_time"] > flow_time])
    return topo_count


def get_tot_inst_count(p_inst_list, flow_time):
    """
    Get the total number of instanton plus anti-instanton.
    """
    inst_count = sum([abs(round(p_inst["estimate_topo_charge"]))
                     for p_inst in p_inst_list if p_inst["flow_time"] > flow_time])
    return inst_count

# --------------------------------------


class InstantonMap:

    """
    self.inst_list
    self.active_inst_list
    #
    self.inst_list[inst_idx] = [
        dict(
            flow_time,
            current_spacing,
            total_site,
            o_xg_d,
            o_xg,
            xg_d,
            xg,
            plaq,
            s_topo_list,
            topo_sphere_sum_radius_list,
            num_plaq,
            dis_sqr_avg,
            dis_sqr_max,
            ),
        ...
    ]
    """

    def __init__(self, *, total_site=None, topo_sphere_sum_radius_list=None, dis_sqr_limit=None, threshold=None):
        """
        total_site = q.Coordinate(get_param(job_tag, "total_site"))
        topo_sphere_sum_radius_list = 5
        dis_sqr_limit = 5
        threshold = 0.1
        """
        self.inst_list = list()
        self.active_inst_list = list()
        self.flow_time = 0.0
        self.step_counter = 0
        self.current_spacing = 1
        self.current_flow_time_multiplier = 1
        self.initial_total_site = total_site
        self.total_site = total_site
        self.geo = q.Geometry(self.total_site)
        self.plaq_xg_arr = mk_plaq_xg_arr(self.geo)
        self.origin_coordinate = q.Coordinate()
        # dis_sqr_limit (always in unit of current lattice spacing)
        # limit for the distance square to identify as the same instanton.
        self.dis_sqr_limit = dis_sqr_limit
        # threshold of (1 - plaq) for instanton detection
        # if `plaq < 1 - threshold`, then this plaq will be considered for instanton tunnelling.
        self.threshold = threshold
        self.topo_sphere_sum_radius_list = topo_sphere_sum_radius_list
        self.info_list = []
        self.flow_time_list = []
        self.tot_topo_list = []

    def shift(self, shift):
        """
        If we shift the field with `shift` by
        `field = field.shift(shift)`
        then we also shift `self.origin_coordinate`,
        which represent the current location of the initial origin
        (in unit of the initial lattice space).
        `shift` is in unit of the current lattice space.
        """
        self.origin_coordinate = self.origin_coordinate + shift * self.current_spacing

    def convert_xg(self, xg):
        """
        Convert current `xg` to the coordinate of the initial lattice.
        """
        assert self.initial_total_site is not None
        if isinstance(xg, q.Coordinate):
            o_xg = (xg * self.current_spacing -
                    self.origin_coordinate) % self.initial_total_site
        elif isinstance(xg, q.CoordinateD):
            initial_total_site_d = q.CoordinateD(self.initial_total_site)
            origin_coordinate_d = q.CoordinateD(self.origin_coordinate)
            o_xg = (xg * self.current_spacing -
                    origin_coordinate_d) % initial_total_site_d
        else:
            assert False
        return o_xg

    @q.timer
    def half_lattice(self):
        assert len(self.active_inst_list) == 0
        self.current_spacing *= 2
        self.current_flow_time_multiplier *= 2**4 / 2
        total_site = q.Coordinate([c // 2 for c in self.total_site])
        assert total_site * 2 == self.total_site
        assert total_site * self.current_spacing == self.initial_total_site
        self.total_site = total_site
        self.geo = q.Geometry(self.total_site)
        self.plaq_xg_arr = mk_plaq_xg_arr(self.geo)

    def acc_time(self, step_size):
        """
        Call after flow.
        """
        self.flow_time += step_size * self.current_flow_time_multiplier
        self.step_counter += 1

    @q.timer
    def acc_topo_info(self, f_plaq, f_topo=None):
        """
        e.g.
        group_extreme_plaq_xg_list = get_group_extreme_plaq_xg_list(extreme_plaq_xg_list, total_site, dis_sqr_limit)
        f_topo = q.gf_topology_field(gf)
        """
        assert self.total_site is not None
        total_site = self.total_site
        total_site_d = q.CoordinateD(total_site)
        extreme_plaq_xg_list = get_extreme_plaq_xg_list(
            self.plaq_xg_arr, f_plaq, self.threshold)
        group_extreme_plaq_xg_list = get_group_extreme_plaq_xg_list(
            extreme_plaq_xg_list, total_site, self.dis_sqr_limit)
        num_group = len(group_extreme_plaq_xg_list)
        if num_group == 0:
            self.inst_list += self.active_inst_list
            self.active_inst_list = []
            return
        for g in group_extreme_plaq_xg_list:
            assert isinstance(g, list)
            assert len(g) > 0
            assert isinstance(g[0], tuple)
            assert len(g[0]) == 3
        plaq_arr = np.array(
            [g[0][0] for g in group_extreme_plaq_xg_list], dtype=np.float64)
        xg_d_arr = np.array(
            [g[0][1] for g in group_extreme_plaq_xg_list], dtype=np.float64)
        dis_sqr_arr_list = [np.array([p[2] for p in g])
                            for g in group_extreme_plaq_xg_list]
        xg_arr = np.array(xg_d_arr, dtype=np.int32)
        if f_topo is not None:
            assert self.topo_sphere_sum_radius_list is not None
            assert isinstance(self.topo_sphere_sum_radius_list, list)
            geo = f_topo.geo
            topo = f_topo.glb_sum()[:].item()
            f_topo_c = q.FieldComplexD(geo, 1)
            f_s_topo = q.FieldRealD(geo, 1)
            f_topo_c[:] = f_topo[:]
            num_radius = len(self.topo_sphere_sum_radius_list)
            s_topo_arr = np.zeros((num_group, num_radius,), dtype=np.float64)
            for idx_radius, radius in enumerate(self.topo_sphere_sum_radius_list):
                f_s_topo_c = q.sphere_sum_field(f_topo_c, radius)
                f_s_topo[:] = f_s_topo_c[:].real
                sp_s_topo = f_s_topo.get_elems_xg(xg_arr)
                s_topo_arr[:, idx_radius] = sp_s_topo[:, 0]
            assert s_topo_arr.shape == (num_group, num_radius,)
        else:
            num_radius = 0
            s_topo_arr = np.array(
                [None for i in range(num_group)], dtype=object)
            assert s_topo_arr.shape == (num_group,)
        assert plaq_arr.shape == (num_group,)
        assert xg_d_arr.shape == (num_group, 4,)
        assert xg_arr.shape == (num_group, 4,)
        assert len(dis_sqr_arr_list) == num_group
        prev_active_inst_list = self.active_inst_list
        self.active_inst_list = []
        for idx in range(num_group):
            xg_d = q.CoordinateD(xg_d_arr[idx])
            xg = q.Coordinate(xg_arr[idx])
            o_xg_d = self.convert_xg(xg_d)
            o_xg = self.convert_xg(xg)
            plaq = plaq_arr[idx].item()
            s_topo = s_topo_arr[idx]
            dis_sqr_arr = dis_sqr_arr_list[idx]
            num_plaq = len(dis_sqr_arr)
            assert dis_sqr_arr.shape == (num_plaq,)
            dis_sqr_avg = np.average(dis_sqr_arr).item()
            dis_sqr_max = np.max(dis_sqr_arr).item()
            if s_topo is not None:
                s_topo = s_topo.tolist()
            info_dict = dict()
            info_dict["flow_time"] = self.flow_time
            info_dict["current_spacing"] = self.current_spacing
            info_dict["total_site"] = self.total_site.to_list()
            info_dict["xg_d"] = xg_d.to_list()
            info_dict["xg"] = xg.to_list()
            info_dict["o_xg"] = o_xg.to_list()
            info_dict["o_xg_d"] = o_xg_d.to_list()
            info_dict["plaq"] = plaq
            info_dict["s_topo"] = s_topo
            info_dict["topo_sphere_sum_radius_list"] = self.topo_sphere_sum_radius_list
            info_dict["num_plaq"] = num_plaq
            info_dict["dis_sqr_avg"] = dis_sqr_avg
            info_dict["dis_sqr_max"] = dis_sqr_max
            is_active = False
            for k in range(len(prev_active_inst_list)):
                inst = prev_active_inst_list[k]
                assert len(inst) > 0
                inst_xg_d = q.CoordinateD(inst[-1]["xg_d"])
                dis_sqr = point_d_dis_sqr(inst_xg_d, xg_d, total_site_d)
                if dis_sqr < self.dis_sqr_limit:
                    is_active = True
                    inst.append(info_dict)
                    self.active_inst_list.append(inst)
                    del prev_active_inst_list[k]
                    break
            if not is_active:
                inst = [info_dict, ]
                self.active_inst_list.append(inst)
        self.inst_list += prev_active_inst_list

# --------------------------------------


@q.timer(is_timer_fork=True)
def compute_inst_map(
    gf,
    *,
    dis_sqr_limit=None,
    threshold=None,
    topo_sphere_sum_radius_list=None,
    plaq_min_threshold=None,
    max_spacing=None,
    flow_step_size_list=None,
    flow_num_step_list=None,
):
    """
    e.g.:
    #
    job_tag = "16IH2"
    set_param(job_tag, "traj_list")(list(range(1000, 4000, 10)))
    set_param(job_tag, "load_config_params")(None)
    #
    get_gf = run_gf(job_tag, traj)
    if get_gf is None:
        return
    gf = get_gf()
    #
    inst_map_obj = compute_inst_map(gf) # Note that `gf` should not have boundary be twisted.
    #
    obj = dict()
    obj["inst_list"] = inst_map.inst_list
    obj["info_list"] = inst_map.info_list
    obj["flow_time_list"] = inst_map.flow_time_list
    obj["tot_topo_list"] = inst_map.tot_topo_list
    """
    fname = q.get_fname()
    if topo_sphere_sum_radius_list is None:
        topo_sphere_sum_radius_list = [2, 3, 4, 5, 6, ]
    q.displayln_info(
        0, f"{fname}: topo_sphere_sum_radius_list={topo_sphere_sum_radius_list}")
    if dis_sqr_limit is None:
        # limit to group plaq to identify as instanton
        dis_sqr_limit = 5
    q.displayln_info(0, f"{fname}: dis_sqr_limit={dis_sqr_limit}")
    if threshold is None:
        # threshold for (1 - plaq) to identify as instanton
        threshold = 0.1
    q.displayln_info(0, f"{fname}: threshold={threshold}")
    if plaq_min_threshold is None:
        # ADJUST ME
        # https://inspirehep.net/literature/166110
        # Tr(1 - U) < 0.03 for SU(2) field
        # suggest 1 - Tr(U)/3 > 0.99 for SU(3) field
        plaq_min_threshold = 0.97**(1/16)
        # plaq_min_threshold = 0.99
    q.displayln_info(0, f"{fname}: plaq_min_threshold={plaq_min_threshold}")
    #
    geo = gf.geo
    total_site = geo.total_site
    #
    if max_spacing is None:
        max_spacing = 1
        total_site_arr = total_site.to_numpy()
        while np.all(total_site_arr % (2 * max_spacing) == 0):
            max_spacing *= 2
    #
    if flow_step_size_list is None:
        flow_step_size_list = [0.05, 0.1, 0.1, ]
    else:
        assert isinstance(flow_step_size_list, list)
        assert len(flow_step_size_list) == 3
    #
    if flow_num_step_list is None:
        flow_num_step_list = [80, 320, 1000000, ]
    else:
        assert isinstance(flow_num_step_list, list)
        assert len(flow_num_step_list) == 3
    #
    inst_map = InstantonMap(
        total_site=total_site,
        topo_sphere_sum_radius_list=topo_sphere_sum_radius_list,
        dis_sqr_limit=dis_sqr_limit,
        threshold=threshold,
    )
    #
    flow_type = None
    plaq = None
    plaq_min = None
    num_instanton = None
    topo = None

    def append_topo_info():
        info = dict()
        info["flow_time"] = inst_map.flow_time
        info["step_counter"] = inst_map.step_counter
        info["flow_type"] = flow_type
        info["current_spacing"] = inst_map.current_spacing
        info["plaq"] = plaq
        info["plaq_min"] = plaq_min
        info["num_instanton"] = num_instanton
        info["topo"] = topo
        inst_map.info_list.append(info)
    #
    flow_type = "Wilson"
    step_size = flow_step_size_list[0]
    num_step = flow_num_step_list[0]
    for i in range(num_step + 1):
        geo = gf.geo
        f_plaq = q.gf_plaq_field(gf)
        plaq_min = np.min(f_plaq.glb_min()[:]).item()
        plaq = f_plaq.glb_sum()[:].sum().item() / geo.total_volume / 6
        num_instanton = ((1 - plaq) * 6 * geo.total_volume) / \
            (8 * np.pi**2 / 6)
        if inst_map.step_counter % 10 == 0:
            f_topo = q.gf_topology_field(gf)
            topo = f_topo.glb_sum()[:].item()
            append_topo_info()
            q.displayln_info(
                0, f"flow_time={inst_map.flow_time:.3f} ; topo", topo)
        q.displayln_info(0, f"{flow_type} ; {geo.total_site.to_list()} ; i={i} ; step_counter={inst_map.step_counter} ; flow_time={inst_map.flow_time:.3f} ; 1-plaq_min={1-plaq_min:.5f} ; 1-plaq={1-plaq:.5f} ; num_instanton={num_instanton:.2f}")
        if i != num_step:
            gf_flow_topo(gf, step_size)
            inst_map.acc_time(step_size)
    #
    if q.is_test():
        q.json_results_append(f"{fname}: {flow_type} plaq", plaq, 1e-10)
        q.json_results_append(f"{fname}: {flow_type} plaq_min", plaq_min, 1e-8)
        q.json_results_append(f"{fname}: {flow_type} topo", topo, 1e-8)
    # q.timer_display()
    #
    flow_type = "Freeze"
    step_size = flow_step_size_list[1]
    num_step = flow_num_step_list[1]
    for i in range(num_step + 1):
        geo = gf.geo
        f_plaq = q.gf_plaq_field(gf)
        plaq_min = np.min(f_plaq.glb_min()[:]).item()
        plaq = f_plaq.glb_sum()[:].sum().item() / geo.total_volume / 6
        num_instanton = ((1 - plaq) * 6 * geo.total_volume) / \
            (8 * np.pi**2 / 6)
        if inst_map.step_counter % 10 == 0:
            f_topo = q.gf_topology_field(gf)
            topo = f_topo.glb_sum()[:].item()
            append_topo_info()
            q.displayln_info(
                0, f"flow_time={inst_map.flow_time:.3f} ; topo", topo)
        q.displayln_info(0, f"{flow_type} ; {geo.total_site.to_list()} ; i={i} ; step_counter={inst_map.step_counter} ; flow_time={inst_map.flow_time:.3f} ; 1-plaq_min={1-plaq_min:.5f} ; 1-plaq={1-plaq:.5f} ; num_instanton={num_instanton:.2f}")
        if plaq_min > 0.995:
            break
        if i != num_step:
            gf_flow_topo(gf, step_size, "Freeze")
            inst_map.acc_time(step_size)
    #
    topo = q.gf_topology(gf)
    inst_map.flow_time_list.append(inst_map.flow_time)
    inst_map.tot_topo_list.append(topo)
    q.displayln_info(
        0, f"flow_time_list={inst_map.flow_time_list} ; tot_topo_list={inst_map.tot_topo_list}")
    if q.is_test():
        q.json_results_append(f"{fname}: {flow_type} plaq", plaq, 1e-10)
        q.json_results_append(f"{fname}: {flow_type} plaq_min", plaq_min, 1e-8)
        q.json_results_append(f"{fname}: {flow_type} topo", topo, 1e-8)
    # q.timer_display()
    #
    flow_type = "Shrink"
    step_size = flow_step_size_list[2]
    num_step = flow_num_step_list[2]
    for i in range(num_step + 1):
        geo = gf.geo
        f_plaq = q.gf_plaq_field(gf)
        plaq_min = np.min(f_plaq.glb_min()[:]).item()
        plaq = f_plaq.glb_sum()[:].sum().item() / geo.total_volume / 6
        num_instanton = ((1 - plaq) * 6 * geo.total_volume) / \
            (8 * np.pi**2 / 6)
        if inst_map.step_counter % 4 == 0:
            f_topo = q.gf_topology_field(gf)
            topo = f_topo.glb_sum()[:].item()
            append_topo_info()
            q.displayln_info(
                0, f"flow_time={inst_map.flow_time:.3f} ; topo", topo)
        else:
            f_topo = None
        q.displayln_info(0, f"{flow_type} ; {geo.total_site.to_list()} ; i={i} ; step_counter={inst_map.step_counter} ; flow_time={inst_map.flow_time:.3f} ; 1-plaq_min={1-plaq_min:.5f} ; 1-plaq={1-plaq:.5f} ; num_instanton={num_instanton:.2f}")
        inst_map.acc_topo_info(f_plaq, f_topo)
        q.displayln_info(
            0, f"num_inst={len(inst_map.inst_list)} ; num_active_inst={len(inst_map.active_inst_list)}")
        if num_instanton < 0.1:
            break
        if abs(topo) < 1e-5 and num_instanton < 0.9:
            break
        if plaq_min > plaq_min_threshold and inst_map.current_spacing * 2 <= max_spacing:
            gf_h = q.gf_reduce_half(gf)
            geo = gf.geo
            geo_h = gf_h.geo
            # gf.show_info()
            # gf_h.show_info()
            f_plaq = q.gf_plaq_field(gf)
            plaq_min = np.min(f_plaq.glb_min()[:]).item()
            plaq = f_plaq.glb_sum()[:].sum().item() / geo.total_volume / 6
            num_instanton = ((1 - plaq) * 6 *
                             geo.total_volume) / (8 * np.pi**2 / 6)
            f_plaq_h = q.gf_plaq_field(gf_h)
            plaq_min_h = np.min(f_plaq_h.glb_min()[:]).item()
            plaq_h = f_plaq_h.glb_sum()[:].sum(
            ).item() / geo_h.total_volume / 6
            num_instanton_h = ((1 - plaq_h) * 6 *
                               geo_h.total_volume) / (8 * np.pi**2 / 6)
            f_topo = q.gf_topology_field(gf)
            topo = f_topo.glb_sum()[:].item()
            f_topo_h = q.gf_topology_field(gf_h)
            topo_h = f_topo_h.glb_sum()[:].item()
            q.displayln_info(
                0, f"===================== Half lattice size ===========================")
            q.displayln_info(0, f"flow_time={inst_map.flow_time}")
            q.displayln_info(0, f"min_plaq", plaq_min, plaq_min_h)
            q.displayln_info(0, f"num_instanton",
                             num_instanton, num_instanton_h)
            q.displayln_info(0, f"topo", topo, topo_h)
            q.displayln_info(
                0, f"===================== Half lattice size ===========================")
            if q.is_test():
                q.json_results_append(
                    f"{fname}: {flow_type} current_spacing={inst_map.current_spacing}")
                q.json_results_append(f"{fname}: {flow_type} plaq", plaq, 1e-8)
                q.json_results_append(
                    f"{fname}: {flow_type} plaq_h", plaq_h, 1e-8)
                q.json_results_append(
                    f"{fname}: {flow_type} plaq_min", plaq_min, 1e-6)
                q.json_results_append(
                    f"{fname}: {flow_type} plaq_min_h", plaq_min_h, 1e-6)
                q.json_results_append(f"{fname}: {flow_type} topo", topo, 1e-4)
                q.json_results_append(
                    f"{fname}: {flow_type} topo_h", topo_h, 1e-4)
            gf = gf_h
            geo = geo_h
            total_site = geo_h.total_site
            inst_map.half_lattice()
            inst_map.flow_time_list.append(inst_map.flow_time)
            inst_map.tot_topo_list.append(topo_h)
            continue
        if i != num_step:
            gf_flow_topo(gf, step_size, "Shrink")
            inst_map.acc_time(step_size)
    obj = dict()
    obj["inst_list"] = inst_map.inst_list
    obj["info_list"] = inst_map.info_list
    obj["flow_time_list"] = inst_map.flow_time_list
    obj["tot_topo_list"] = inst_map.tot_topo_list
    q.displayln_info(0, f"tot_topo_list={inst_map.tot_topo_list}")
    return obj


@q.timer
def displayln_info_inst_map_obj(inst_map_obj):
    """
    e.g.:
    inst_map_obj = compute_inst_map(gf)
    displayln_info_inst_map_obj(inst_map_obj)
    """
    fname = q.get_fname()
    obj = inst_map_obj
    #
    q.displayln_info(0, f"{fname}: {obj.keys()}")
    #
    info_list = obj["info_list"]
    inst_list = obj["inst_list"]
    tot_topo_list = obj["tot_topo_list"]
    flow_time_list = obj["flow_time_list"]
    #
    p_inst_list = process_inst_list(inst_list)
    #
    if q.is_test():
        for p_inst in p_inst_list:
            inst_idx = p_inst["inst_idx"]
            o_xg_d = p_inst["o_xg_d"]
            o_total_site = p_inst["o_total_site"]
            current_spacing = p_inst["current_spacing"]
            plaq = p_inst["plaq"]
            flow_time = p_inst["flow_time"]
            delta_s_topo = p_inst["delta_s_topo"]
            estimate_topo_charge = p_inst["estimate_topo_charge"]
            q.json_results_append(
                f"inst_idx={inst_idx} ; o_xg_d={o_xg_d} ; o_total_site={o_total_site} ; current_spacing={current_spacing}")
            q.json_results_append(f"plaq", plaq, 1e-6)
            q.json_results_append(f"flow_time", flow_time, 1e-3)
            q.json_results_append(f"delta_s_topo", np.array(
                delta_s_topo, dtype=np.float64), 1e-6)
            q.json_results_append(f"estimate_topo_charge",
                                  estimate_topo_charge, 1e-4)
    #
    s_p_inst_list = sorted(p_inst_list, key=lambda v: -v["flow_time"])
    #
    q.displayln_info(0, f"")
    for idx, (flow_time, tot_topo,) in enumerate(zip(flow_time_list, tot_topo_list)):
        topo_count = get_tot_topo_count(p_inst_list, flow_time)
        inst_count = get_tot_inst_count(p_inst_list, flow_time)
        current_spacing = 2**idx
        q.displayln_info(
            0, f"{idx:3}: flow_time={flow_time:9.2f} ; tot_topo={tot_topo:10.3f} ; topo_count={topo_count:4} ; inst_count={inst_count:4} ; current_spacing={current_spacing}")
    #
    q.displayln_info(0, f"")
    #
    for topo_info in info_list:
        flow_type = topo_info["flow_type"]
        flow_time = topo_info["flow_time"]
        while len(s_p_inst_list) > 0:
            if s_p_inst_list[-1]["flow_time"] <= flow_time:
                displayln_info_p_inst(s_p_inst_list.pop())
            else:
                break
        displayln_info_topo_info(topo_info)
        topo_count = get_tot_topo_count(p_inst_list, flow_time)
        inst_count = get_tot_inst_count(p_inst_list, flow_time)
        q.displayln_info(
            0, f"{flow_type} flow_time={flow_time:9.2f} ; topo_count={topo_count:4} ; inst_count={inst_count:4}")
    #
    q.displayln_info(0, f"")
    #
    tot_topo_direct = round(tot_topo_list[0])
    tot_topo_count = get_tot_topo_count(p_inst_list, 0)
    tot_inst_count = get_tot_inst_count(p_inst_list, 0)
    #
    q.displayln_info(
        0, f"{fname}: tot_topo_direct={tot_topo_direct} ; tot_topo_count={tot_topo_count}")
    if tot_topo_direct == tot_topo_count:
        q.displayln_info(0, f"{fname}: Topological charge matched.")
    else:
        q.displayln_info(-1,
                         f"{fname}: WARNING: Topological charge does not match.")


@q.timer(is_timer_fork=True)
def smear_measure_topo(
    gf,
    smear_info_list=None,
    *,
    energy_derivative_info=None,
    info_path=None,
    density_field_path=None,
    is_show_topo_terms=False,
):
    r"""
    return topo_list, energy_list,
    gf will be modified in-place.
    ```
    smear_info_list = [
        [
            step_size,
            n_step,
            flow_type="Wilson",
            wilson_flow_integrator_type="euler",
        ],
        ...
    ]
    energy_derivative_info = [
        step_size,
        flow_type="Wilson",
        wilson_flow_integrator_type="euler",
    ]
    flow_type in [
        None,
        "Wilson", # c1 = 0.0
        "Iwasaki", # c1 = -0.331
        "DBW2", # c1 = -1.4008
        "Shrink",
        "Freeze",
        "Localize",
        c1,
    ]
    wilson_flow_integrator_type in [
        None,
        "euler",
        "runge-kutta",
    ]
    ```
    ``info_path`` can be the same as ``density_field_path``.
    """
    fname = q.get_fname()
    if smear_info_list is None:
        smear_info_list = [
            [0.05, 20, "Wilson", "euler", ],
            [0.05, 20, "Wilson", "euler", ],
            [0.05, 20, "Wilson", "euler", ],
            [0.01, 50, "DBW2", "euler", ],
            [0.01, 50, "DBW2", "euler", ],
            [0.01, 50, "DBW2", "euler", ],
            [0.01, 50, "DBW2", "euler", ],
        ]
    if energy_derivative_info is None:
        energy_derivative_info = [0.0125, 0.0, "euler", ]
    q.displayln_info(0, f"{fname}: smear_info_list =")
    q.displayln_info(0, pformat(smear_info_list))
    geo = gf.geo
    total_volume = geo.total_volume
    total_site = geo.total_site
    spatial_volume = total_volume / total_site[3]
    flow_time = 0
    topo_list = []
    energy_list = []

    @q.timer
    def smear(
        step_size,
        n_step,
        flow_type="Wilson",
        wilson_flow_integrator_type="euler",
    ):
        nonlocal flow_time
        flow_time_target = flow_time + n_step * step_size
        while flow_time < flow_time_target - step_size / 3:
            flow_time += gf_flow_topo(gf, step_size, flow_type, wilson_flow_integrator_type)
            plaq = gf.plaq()
            f_plaq = q.gf_plaq_field(gf)
            plaq_min = np.min(f_plaq.glb_min()[:]).item()
            plaq_max = np.max(f_plaq.glb_max()[:]).item()
            energy_density_field = q.gf_energy_density_field(gf)
            energy_density = energy_density_field.glb_sum()[
                :].item() / total_volume
            energy_density_tslice = (energy_density_field.glb_sum_tslice()[
                                     :].ravel() / spatial_volume).tolist()
            energy_list.append({
                "flow_time": flow_time,
                "plaq": plaq,
                "plaq_min": plaq_min,
                "plaq_max": plaq_max,
                "energy_density": energy_density,
                "energy_density_tslice": energy_density_tslice,
            })

    @q.timer
    def measure():
        gf.show_info()
        plaq = gf.plaq()
        f_plaq = q.gf_plaq_field(gf)
        plaq_min = np.min(f_plaq.glb_min()[:]).item()
        plaq_max = np.max(f_plaq.glb_max()[:]).item()
        plaq_action_density_field = q.gf_plaq_action_density_field(gf)
        plaq_action_density = plaq_action_density_field.glb_sum()[
            :].item() / total_volume
        plaq_action_density_tslice = (plaq_action_density_field.glb_sum_tslice()[
                                      :].ravel() / spatial_volume).tolist()
        energy_density_field = q.gf_energy_density_field(gf)
        energy_density = energy_density_field.glb_sum()[
            :].item() / total_volume
        energy_density_tslice = (energy_density_field.glb_sum_tslice()[
                                 :].ravel() / spatial_volume).tolist()
        topo_field_clf = q.gf_topology_field_clf(gf)
        topo_clf = topo_field_clf.glb_sum()[:].item()
        topo_clf_tslice = topo_field_clf.glb_sum_tslice()[:].ravel().tolist()
        topo_field = q.gf_topology_field(gf)
        topo = topo_field.glb_sum()[:].item()
        topo_tslice = topo_field.glb_sum_tslice()[:].ravel().tolist()
        abs_topo_field = topo_field.copy()
        abs_topo_field[:] = np.abs(topo_field[:])
        abs_topo = abs_topo_field.glb_sum()[:].item()
        abs_topo_tslice = abs_topo_field.glb_sum_tslice()[:].ravel().tolist()
        energy_deriv_density_field = gf_energy_derivative_density_field_topo(
            gf,
            epsilon=energy_derivative_info[0],
            flow_typw=energy_derivative_info[1],
            wilson_flow_integrator_type=energy_derivative_info[2],
        )
        energy_deriv = energy_deriv_density_field.glb_sum()[
            :].item() / total_volume
        energy_deriv_tslice = (energy_deriv_density_field.glb_sum_tslice()[
                               :].ravel() / spatial_volume).tolist()
        q.displayln_info(
            0, f"{fname}: t={flow_time} ; energy_density={energy_density} ; energy_deriv={energy_deriv} ; topo={topo}")
        q.displayln_info(0, pformat(
            list(enumerate(zip(energy_density_tslice, energy_deriv_tslice, topo_tslice)))))
        if is_show_topo_terms:
            topo_terms = q.gf_topology_terms(gf)
            q.displayln_info(
                0, f"{fname}: t={flow_time} ; topo={topo} ; {sum(topo_terms)}")
            topo_terms_str = ',\n  '.join([str(x) for x in topo_terms])
            q.displayln_info(0, f"[ {topo_terms_str},\n]")
        if density_field_path is not None:
            energy_density_field.save_double(
                f"{density_field_path}/smear-step-{len(topo_list)}/energy_density.field")
            topo_field.save_double(
                f"{density_field_path}/smear-step-{len(topo_list)}/topo.field")
            energy_deriv_density_field.save_double(
                f"{density_field_path}/smear-step-{len(topo_list)}/energy_deriv_density_field.field")
        topo_list.append({
            "flow_time": flow_time,
            "plaq": plaq,
            "plaq_min": plaq_min,
            "plaq_max": plaq_max,
            "energy_density": energy_density,
            "energy_density_tslice": energy_density_tslice,
            "energy_deriv": energy_deriv,
            "energy_deriv_tslice": energy_deriv_tslice,
            "topo": topo,
            "topo_tslice": topo_tslice,
            "topo_clf": topo_clf,
            "topo_clf_tslice": topo_clf_tslice,
            "abs_topo": abs_topo,
            "abs_topo_tslice": abs_topo_tslice,
            "plaq_action_density": plaq_action_density,
            "plaq_action_density_tslice": plaq_action_density_tslice,
        })

    measure()
    for si in smear_info_list:
        smear(*si)
        measure()
    if info_path is not None:
        q.save_json_obj(smear_info_list, f"{info_path}/smear-info-list.json", indent=2)
        q.save_pickle_obj(topo_list, f"{info_path}/info.pickle")
        q.save_pickle_obj(energy_list, f"{info_path}/energy-list.pickle")
    return topo_list, energy_list,


# --------------------------------------



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils.all cimport *
from . cimport everything as cc
from .field_types cimport FieldRealD
from .geometry cimport Geometry
from .qcd cimport GaugeField
from .hmc cimport GaugeMomentum
from .gauge_action cimport GaugeAction

from .hmc import set_gm_force, gf_evolve
from .wilson_flow import gf_wilson_flow_step, gf_energy_density_field, gf_energy_derivative_density_field
import qlat_utils as q

@q.timer
def gf_plaq_action_density_field(GaugeField gf):
    r"""
    return paf
    paf.geo.multiplicity == 1
    \sum_P (1 - 1/3 * Re Tr U_P)
    #
    6 plaqs in the above sum (for each site).
    #
    Action = beta * total_volume * 6 * (1 - avg_plaq)
    Action = beta * total_volume * action_density
    Single instanton action = 8 * sqr(PI) / g^2
    beta = 6/g^2
    """
    paf = FieldRealD()
    cc.clf_plaq_action_density_field(paf.xx, gf.xxx().val())
    return paf

@q.timer
def gf_spatial_plaq_action_density_field(GaugeField gf):
    r"""
    return paf
    paf.geo.multiplicity == 1
    \sum_P(spatial only) (1 - 1/3 * Re Tr U_P)
    """
    paf = FieldRealD()
    cc.clf_spatial_plaq_action_density_field(paf.xx, gf.xxx().val())
    return paf

@q.timer
def gf_plaq_action_density(GaugeField gf):
    r"""
    return pa
    ininstance(pa, float)
    pa = gf_plaq_action_density_field(gf).glb_sum()[:].item() / total_volume
    """
    cdef Geometry geo = gf.geo
    cdef cc.Long total_volume = geo.total_volume
    return gf_plaq_action_density_field(gf).glb_sum()[:].item() / total_volume

@q.timer
def gf_spatial_plaq_action_density(GaugeField gf):
    r"""
    return pa
    ininstance(pa, float)
    pa = gf_spatial_plaq_action_density_field(gf).glb_sum()[:].item() / total_volume
    """
    cdef Geometry geo = gf.geo
    cdef cc.Long total_volume = geo.total_volume
    return gf_spatial_plaq_action_density_field(gf).glb_sum()[:].item() / total_volume

@q.timer
def gf_topology_field_clf(GaugeField gf):
    r"""
    return topf
    topf.geo.multiplicity == 1
    Use the basic gf_clover_leaf_field
    NOT using 5 loop improved definition
    """
    topf = FieldRealD()
    cc.clf_topology_field(topf.xx, gf.xxx().val())
    return topf

@q.timer
def gf_topology_clf(GaugeField gf):
    r"""
    return top
    ininstance(top, float)
    Use the basic gf_clover_leaf_field
    NOT using 5 loop improved definition
    """
    return gf_topology_field_clf(gf).glb_sum()[:].item()

@q.timer
def gf_topology_field(GaugeField gf):
    r"""
    return topf
    topf.geo.multiplicity == 1
    Using the 5 loop improved definition
    https://arxiv.org/pdf/hep-lat/9701012v2.pdf
    """
    topf = FieldRealD()
    cc.clf_topology_field_5(topf.xx, gf.xxx().val())
    return topf

@q.timer
def gf_topology(GaugeField gf):
    r"""
    return top
    ininstance(top, float)
    Using the 5 loop improved definition Eq. (2-7)
    https://arxiv.org/pdf/hep-lat/9701012v2.pdf
    """
    return gf_topology_field(gf).glb_sum()[:].item()

@q.timer
def gf_topology_terms_field(GaugeField gf):
    r"""
    return topf;
    topf.geo.multiplicity == 5
    sum of the 5 terms should equal to gf_topology_field
    """
    topf = FieldRealD()
    cc.clf_topology_field_5_terms(topf.xx, gf.xxx().val())
    return topf

@q.timer
def gf_topology_terms(GaugeField gf):
    r"""
    return top_terms;
    top_terms.shape == (5,)
    top_terms.dtype == np.float64
    sum of the 5 terms should equal to gf_topology
    """
    return gf_topology_terms_field(gf).glb_sum()[0, :]



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils.all cimport *
from . cimport everything as cc
from .geometry cimport Geometry
from .gauge_action cimport GaugeAction
from .field_types cimport (
        FieldColorMatrix,
        FieldRealD,
        )
from .qcd cimport GaugeField

import cqlat as c
import qlat_utils as q
import numpy as np
from .mpi import glb_sum_double

import math

cdef class GaugeMomentum(FieldColorMatrix):

    def __init__(self, Geometry geo=None):
        super().__init__(geo, 4)

    cdef cc.Handle[cc.GaugeMomentum] xxx(self):
        assert self.xx.multiplicity == 4 or self.xx.multiplicity == 0
        return cc.Handle[cc.GaugeMomentum](<cc.GaugeMomentum&>self.xx)

    def set_rand(self, RngState rng, cc.RealD sigma=1.0):
        set_rand_gauge_momentum(self, sigma, rng)

    def set_rand_fa(self, FieldRealD mf, RngState rng):
        set_rand_gauge_momentum_fa(self, mf, rng)

###

def set_rand_gauge_momentum(GaugeMomentum gm, cc.RealD sigma, RngState rng):
    return cc.set_rand_gauge_momentum(gm.xxx().val(), sigma, rng.xx)

def set_rand_gauge_momentum_fa(GaugeMomentum gm, FieldRealD mf, RngState rng):
    return cc.set_rand_gauge_momentum(gm.xxx().val(), mf.xx, rng.xx)

def gm_hamilton_node(GaugeMomentum gm):
    return cc.gm_hamilton_node(gm.xxx().val())

def gm_hamilton_node_fa(GaugeMomentum gm, FieldRealD mf):
    return cc.gm_hamilton_node(gm.xxx().val(), mf.xx)

def gf_hamilton_node(GaugeField gf, GaugeAction ga):
    return cc.gf_hamilton_node(gf.xxx().val(), ga.xx)

def gf_evolve(GaugeField gf, GaugeMomentum gm, cc.RealD step_size):
    return cc.gf_evolve(gf.xxx().val(), gm.xxx().val(), step_size)

def gf_evolve_dual(GaugeField gf, GaugeMomentum gm_dual, cc.RealD step_size):
    return cc.gf_evolve_dual(gf.xxx().val(), gm_dual.xxx().val(), step_size)

def gf_evolve_fa(GaugeField gf, GaugeMomentum gm, FieldRealD mf, cc.RealD step_size):
    return cc.gf_evolve(gf.xxx().val(), gm.xxx().val(), mf.xx, step_size)

def gf_evolve_fa_dual(GaugeField gf, GaugeMomentum gm_dual, FieldRealD mf_dual, cc.RealD step_size):
    return cc.gf_evolve_dual(gf.xxx().val(), gm_dual.xxx().val(), mf_dual.xx, step_size)

def field_color_matrix_exp(FieldColorMatrix fc1, cc.PyComplexD coef):
    cdef FieldColorMatrix fc = FieldColorMatrix()
    cdef cc.ComplexD cc_coef = cc.ccpy_d(coef)
    cc.field_color_matrix_exp(fc.xx, fc1.xx, cc_coef)
    return fc

def field_color_matrix_mul(FieldColorMatrix fc1, FieldColorMatrix fc2):
    cdef FieldColorMatrix fc = FieldColorMatrix()
    cc.field_color_matrix_mul(fc.xx, fc1.xx, fc2.xx)
    return fc

def set_gm_force(GaugeMomentum gm_force, GaugeField gf, GaugeAction ga):
    return cc.set_gm_force(gm_force.xxx().val(), gf.xxx().val(), ga.xx)

def set_gm_force_dual(GaugeMomentum gm_force_dual, GaugeField gf, GaugeMomentum gm_force):
    return cc.set_gm_force_dual(gm_force_dual.xxx().val(), gf.xxx().val(), gm_force.xxx().val())

def project_gauge_transform(GaugeMomentum gm, GaugeMomentum gm_dual, FieldRealD mf, FieldRealD mf_dual):
    """
    Project out the pure gauge transformation movement.
    """
    return cc.project_gauge_transform(gm.xxx().val(), gm_dual.xxx().val(), mf.xx, mf_dual.xx)

def set_gauge_transform_momentum(GaugeMomentum gm, GaugeMomentum gm_dual, FieldColorMatrix gtm):
    """
    Set `gm` and `gm_dual` with `gtm`.
    The overall effects of `gm` and `gm_dual` with `gf_evolve` and `gf_evolve_dual` is a pure gauge transformation from `gtm`.
    The name `gtm` stands for gauge tranformation momentum.
    """
    return cc.set_gauge_transform_momentum(gm.xxx().val(), gm_dual.xxx().val(), gtm.xx)

def dot_gauge_momentum(GaugeMomentum gm1, GaugeMomentum gm2):
    """
    return RealD dot field with multiplicity = 4
    """
    cdef FieldRealD f = FieldRealD()
    cc.dot_gauge_momentum(f.xx, gm1.xxx().val(), gm2.xxx().val())
    return f

def set_anti_hermitian_matrix_from_basis(FieldColorMatrix fc, FieldRealD basis):
    r"""
    fc[x, m] = \sum_a T_a * basis[x, m * 8 + a]
    T_a^dagger = -T_a
    Tr[T_a T_b] = -2 \delta_{a,b}
    """
    cc.set_anti_hermitian_matrix_from_basis(fc.xx, basis.xx)

def set_basis_from_anti_hermitian_matrix(FieldRealD basis, FieldColorMatrix fc):
    r"""
    fc[x, m] = \sum_a T_a * basis[x, m * 8 + a]
    T_a^dagger = -T_a
    Tr[T_a T_b] = -2 \delta_{a,b}
    """
    cc.set_basis_from_anti_hermitian_matrix(basis.xx, fc.xx)

@q.timer_verbose
def metropolis_accept(delta_h, traj, rs):
    flag_d = 0.0
    accept_prob = 0.0
    if q.get_id_node() == 0:
        if delta_h <= 0.0:
            accept_prob = 1.0
            flag_d = 1.0
        else:
            accept_prob = math.exp(-delta_h)
            rand_num = rs.u_rand_gen(1.0, 0.0)
            if rand_num <= accept_prob:
                flag_d = 1.0
    flag_d = glb_sum_double(flag_d)
    accept_prob = glb_sum_double(accept_prob)
    flag = flag_d > 0.5
    q.displayln_info(f"metropolis_accept: flag={flag:d} with accept_prob={accept_prob * 100.0:.1f}% delta_h={delta_h:.16f} traj={traj}")
    return flag, accept_prob

@q.timer
def gm_evolve_fg_pure_gauge(gm, gf_init, ga, fg_dt, dt):
    geo = gf_init.geo
    gf = GaugeField(geo)
    gf @= gf_init
    gm_force = GaugeMomentum(geo)
    set_gm_force(gm_force, gf, ga)
    gf_evolve(gf, gm_force, fg_dt)
    set_gm_force(gm_force, gf, ga)
    gm_force *= dt
    gm += gm_force

@q.timer(is_timer_fork=True)
def run_hmc_evolve_pure_gauge(gm, gf, ga, rs, n_step, md_time=1.0):
    energy = gm_hamilton_node(gm) + gf_hamilton_node(gf, ga)
    dt = md_time / n_step
    lam = 0.5 * (1.0 - 1.0 / math.sqrt(3.0))
    theta = (2.0 - math.sqrt(3.0)) / 48.0
    ttheta = theta * dt * dt * dt
    gf_evolve(gf, gm, lam * dt)
    for i in range(n_step):
        gm_evolve_fg_pure_gauge(gm, gf, ga, 4.0 * ttheta / dt, 0.5 * dt)
        gf_evolve(gf, gm, (1.0 - 2.0 * lam) * dt)
        gm_evolve_fg_pure_gauge(gm, gf, ga, 4.0 * ttheta / dt, 0.5 * dt)
        if i < n_step - 1:
            gf_evolve(gf, gm, 2.0 * lam * dt)
        else:
            gf_evolve(gf, gm, lam * dt)
    gf.unitarize()
    delta_h = gm_hamilton_node(gm) + gf_hamilton_node(gf, ga) - energy
    delta_h = glb_sum_double(delta_h)
    return delta_h

@q.timer(is_timer_fork=True)
def run_hmc_pure_gauge(gf, ga, traj, rs, *, is_reverse_test=False, n_step=6, md_time=1.0, is_always_accept=False):
    fname = q.get_fname()
    rs = rs.split(f"{traj}")
    geo = gf.geo
    gf0 = GaugeField(geo)
    gf0 @= gf
    gm = GaugeMomentum(geo)
    gm.set_rand(rs.split("set_rand_gauge_momentum"), 1.0)
    delta_h = run_hmc_evolve_pure_gauge(gm, gf0, ga, rs, n_step, md_time)
    if is_reverse_test:
        gm_r = GaugeMomentum(geo)
        gm_r @= gm
        gf0_r = GaugeField(geo)
        gf0_r @= gf0
        delta_h_rev = run_hmc_evolve_pure_gauge(gm_r, gf0_r, ga, rs, n_step, -md_time)
        gf0_r -= gf
        q.displayln_info(f"{fname}: reversed delta_diff: {delta_h + delta_h_rev} / {delta_h}")
        gf_diff_norm = q.qnorm(gf0_r)
        gf_norm = q.qnorm(gf0)
        q.displayln_info(f"{fname}: reversed gf_diff: {gf_diff_norm} / {gf_norm}")
        assert gf_diff_norm <= 1e-12 * gf_norm
    flag, accept_prob = metropolis_accept(delta_h, traj, rs.split("metropolis_accept"))
    if flag or is_always_accept:
        q.displayln_info(f"{fname}: update gf (traj={traj})")
        gf @= gf0
    return flag, delta_h
